@@ -19,8 +19,8 @@
 #include "common/clock.h"
 #include "common/ringbuffer/frontend_types.h"
 
-#define LTTNG_COMPACT_EVENT_BITS       5
-#define LTTNG_COMPACT_TSC_BITS         27
+#define LTTNG_COMPACT_EVENT_BITS	5
+#define LTTNG_COMPACT_TIMESTAMP_BITS	27
 
 /*
  * Keep the natural field alignment for _each field_ within this structure if
@@ -51,7 +51,9 @@ struct packet_header {
 						 * the beginning of the trace.
 						 * (may overflow)
 						 */
+#ifdef RING_BUFFER_CLIENT_HAS_CPU_ID
 		uint32_t cpu_id;		/* CPU id associated with stream */
+#endif
 		uint8_t header_end;		/* End of header */
 	} ctx;
 };
@@ -60,7 +62,6 @@ struct lttng_client_ctx {
 	size_t packet_context_len;
 	size_t event_context_len;
 	struct lttng_ust_ctx *chan_ctx;
-	struct lttng_ust_ctx *event_ctx;
 };
 
 /*
@@ -156,7 +157,7 @@ size_t record_header_size(
 	case 1:	/* compact */
 		padding = lttng_ust_ring_buffer_align(offset, lttng_ust_rb_alignof(uint32_t));
 		offset += padding;
-		if (!(ctx->priv->rflags & (RING_BUFFER_RFLAG_FULL_TSC | LTTNG_RFLAG_EXTENDED))) {
+		if (!(ctx->priv->rflags & (RING_BUFFER_RFLAG_FULL_TIMESTAMP | LTTNG_RFLAG_EXTENDED))) {
 			offset += sizeof(uint32_t);	/* id and timestamp */
 		} else {
 			/* Minimum space taken by LTTNG_COMPACT_EVENT_BITS id */
@@ -172,7 +173,7 @@ size_t record_header_size(
 		padding = lttng_ust_ring_buffer_align(offset, lttng_ust_rb_alignof(uint16_t));
 		offset += padding;
 		offset += sizeof(uint16_t);
-		if (!(ctx->priv->rflags & (RING_BUFFER_RFLAG_FULL_TSC | LTTNG_RFLAG_EXTENDED))) {
+		if (!(ctx->priv->rflags & (RING_BUFFER_RFLAG_FULL_TIMESTAMP | LTTNG_RFLAG_EXTENDED))) {
 			offset += lttng_ust_ring_buffer_align(offset, lttng_ust_rb_alignof(uint32_t));
 			offset += sizeof(uint32_t);	/* timestamp */
 		} else {
@@ -189,8 +190,6 @@ size_t record_header_size(
 	}
 	offset += ctx_get_aligned_size(offset, client_ctx->chan_ctx,
 			client_ctx->packet_context_len);
-	offset += ctx_get_aligned_size(offset, client_ctx->event_ctx,
-			client_ctx->event_context_len);
 	*pre_header_padding = padding;
 	return offset - orig_offset;
 }
@@ -235,14 +234,14 @@ void lttng_write_event_header(const struct lttng_ust_ring_buffer_config *config,
 				event_id);
 		bt_bitfield_write(&id_time, uint32_t,
 				LTTNG_COMPACT_EVENT_BITS,
-				LTTNG_COMPACT_TSC_BITS,
-				ctx->priv->tsc);
+				LTTNG_COMPACT_TIMESTAMP_BITS,
+				ctx->priv->timestamp);
 		lib_ring_buffer_write(config, ctx, &id_time, sizeof(id_time));
 		break;
 	}
 	case 2:	/* large */
 	{
-		uint32_t timestamp = (uint32_t) ctx->priv->tsc;
+		uint32_t timestamp = (uint32_t) ctx->priv->timestamp;
 		uint16_t id = event_id;
 
 		lib_ring_buffer_write(config, ctx, &id, sizeof(id));
@@ -255,7 +254,6 @@ void lttng_write_event_header(const struct lttng_ust_ring_buffer_config *config,
 	}
 
 	ctx_record(ctx, lttng_chan, client_ctx->chan_ctx);
-	ctx_record(ctx, lttng_chan, client_ctx->event_ctx);
 	lttng_ust_ring_buffer_align_ctx(ctx, ctx->largest_align);
 
 	return;
@@ -275,7 +273,7 @@ void lttng_write_event_header_slow(const struct lttng_ust_ring_buffer_config *co
 
 	switch (lttng_chan->priv->header_type) {
 	case 1:	/* compact */
-		if (!(ctx_private->rflags & (RING_BUFFER_RFLAG_FULL_TSC | LTTNG_RFLAG_EXTENDED))) {
+		if (!(ctx_private->rflags & (RING_BUFFER_RFLAG_FULL_TIMESTAMP | LTTNG_RFLAG_EXTENDED))) {
 			uint32_t id_time = 0;
 
 			bt_bitfield_write(&id_time, uint32_t,
@@ -284,12 +282,12 @@ void lttng_write_event_header_slow(const struct lttng_ust_ring_buffer_config *co
 					event_id);
 			bt_bitfield_write(&id_time, uint32_t,
 					LTTNG_COMPACT_EVENT_BITS,
-					LTTNG_COMPACT_TSC_BITS,
-					ctx_private->tsc);
+					LTTNG_COMPACT_TIMESTAMP_BITS,
+					ctx_private->timestamp);
 			lib_ring_buffer_write(config, ctx, &id_time, sizeof(id_time));
 		} else {
 			uint8_t id = 0;
-			uint64_t timestamp = ctx_private->tsc;
+			uint64_t timestamp = ctx_private->timestamp;
 
 			bt_bitfield_write(&id, uint8_t,
 					0,
@@ -305,8 +303,8 @@ void lttng_write_event_header_slow(const struct lttng_ust_ring_buffer_config *co
 		break;
 	case 2:	/* large */
 	{
-		if (!(ctx_private->rflags & (RING_BUFFER_RFLAG_FULL_TSC | LTTNG_RFLAG_EXTENDED))) {
-			uint32_t timestamp = (uint32_t) ctx_private->tsc;
+		if (!(ctx_private->rflags & (RING_BUFFER_RFLAG_FULL_TIMESTAMP | LTTNG_RFLAG_EXTENDED))) {
+			uint32_t timestamp = (uint32_t) ctx_private->timestamp;
 			uint16_t id = event_id;
 
 			lib_ring_buffer_write(config, ctx, &id, sizeof(id));
@@ -314,7 +312,7 @@ void lttng_write_event_header_slow(const struct lttng_ust_ring_buffer_config *co
 			lib_ring_buffer_write(config, ctx, &timestamp, sizeof(timestamp));
 		} else {
 			uint16_t id = 65535;
-			uint64_t timestamp = ctx_private->tsc;
+			uint64_t timestamp = ctx_private->timestamp;
 
 			lib_ring_buffer_write(config, ctx, &id, sizeof(id));
 			/* Align extended struct on largest member */
@@ -329,7 +327,6 @@ void lttng_write_event_header_slow(const struct lttng_ust_ring_buffer_config *co
 		WARN_ON_ONCE(1);
 	}
 	ctx_record(ctx, lttng_chan, client_ctx->chan_ctx);
-	ctx_record(ctx, lttng_chan, client_ctx->event_ctx);
 	lttng_ust_ring_buffer_align_ctx(ctx, ctx->largest_align);
 }
 
@@ -364,7 +361,7 @@ static size_t client_packet_header_size(void)
 	return offsetof(struct packet_header, ctx.header_end);
 }
 
-static void client_buffer_begin(struct lttng_ust_ring_buffer *buf, uint64_t tsc,
+static void client_buffer_begin(struct lttng_ust_ring_buffer *buf, uint64_t timestamp,
 				unsigned int subbuf_idx,
 				struct lttng_ust_shm_handle *handle)
 {
@@ -383,46 +380,55 @@ static void client_buffer_begin(struct lttng_ust_ring_buffer *buf, uint64_t tsc,
 	header->magic = CTF_MAGIC_NUMBER;
 	memcpy(header->uuid, lttng_chan->priv->uuid, sizeof(lttng_chan->priv->uuid));
 	header->stream_id = lttng_chan->priv->id;
+#ifdef RING_BUFFER_CLIENT_HAS_CPU_ID
 	header->stream_instance_id = buf->backend.cpu;
-	header->ctx.timestamp_begin = tsc;
+#else
+	header->stream_instance_id = 0;
+#endif
+	header->ctx.timestamp_begin = timestamp;
 	header->ctx.timestamp_end = 0;
 	header->ctx.content_size = ~0ULL; /* for debugging */
 	header->ctx.packet_size = ~0ULL;
 	header->ctx.packet_seq_num = chan->backend.num_subbuf * cnt + subbuf_idx;
 	header->ctx.events_discarded = 0;
+#ifdef RING_BUFFER_CLIENT_HAS_CPU_ID
 	header->ctx.cpu_id = buf->backend.cpu;
+#endif
 }
 
 /*
  * offset is assumed to never be 0 here : never deliver a completely empty
  * subbuffer. data_size is between 1 and subbuf_size.
  */
-static void client_buffer_end(struct lttng_ust_ring_buffer *buf, uint64_t tsc,
+static void client_buffer_end(struct lttng_ust_ring_buffer *buf, uint64_t timestamp,
 			      unsigned int subbuf_idx, unsigned long data_size,
 			      struct lttng_ust_shm_handle *handle,
-			      const struct lttng_ust_ring_buffer_ctx *ctx)
+			      const struct lttng_ust_ring_buffer_ctx *ctx __attribute__((unused)))
 {
 	struct lttng_ust_ring_buffer_channel *chan = shmp(handle, buf->backend.chan);
+	struct commit_counters_cold *cc_cold = shmp_index(handle, buf->commit_cold, subbuf_idx);
 	struct packet_header *header =
 		(struct packet_header *)
 			lib_ring_buffer_offset_address(&buf->backend,
 				subbuf_idx * chan->backend.subbuf_size,
 				handle);
-	unsigned long records_lost = 0;
+	ssize_t page_size = LTTNG_UST_PAGE_SIZE;
 
 	assert(header);
+	assert(cc_cold);
 	if (!header)
 		return;
-	header->ctx.timestamp_end = tsc;
+	if (page_size < 0)
+		return;
+	if (!cc_cold)
+		return;
+
+	header->ctx.timestamp_end = timestamp;
 	header->ctx.content_size =
 		(uint64_t) data_size * CHAR_BIT;		/* in bits */
 	header->ctx.packet_size =
-		(uint64_t) LTTNG_UST_PAGE_ALIGN(data_size) * CHAR_BIT;	/* in bits */
-
-	records_lost += lib_ring_buffer_get_records_lost_full(&client_config, ctx);
-	records_lost += lib_ring_buffer_get_records_lost_wrap(&client_config, ctx);
-	records_lost += lib_ring_buffer_get_records_lost_big(&client_config, ctx);
-	header->ctx.events_discarded = records_lost;
+		(uint64_t) LTTNG_UST_ALIGN(data_size, page_size) * CHAR_BIT;	/* in bits */
+	header->ctx.events_discarded = cc_cold->end_events_discarded;
 }
 
 static int client_buffer_create(
@@ -546,6 +552,27 @@ static int client_stream_id(struct lttng_ust_ring_buffer *buf __attribute__((unu
 	return 0;
 }
 
+/*
+ * Samples the events discarded counters of the ring buffer.
+ */
+static int client_current_events_discarded(
+		struct lttng_ust_ring_buffer *buf,
+		struct lttng_ust_ring_buffer_channel *chan,
+		uint64_t *events_discarded)
+{
+	uint64_t e = 0;
+
+	if (!events_discarded) {
+		return -1;
+	}
+
+	e += v_read(&chan->backend.config, &buf->records_lost_full);
+	e += v_read(&chan->backend.config, &buf->records_lost_wrap);
+	e += v_read(&chan->backend.config, &buf->records_lost_big);
+	*events_discarded = e;
+	return 0;
+}
+
 static int client_current_timestamp(
 		struct lttng_ust_ring_buffer *buf  __attribute__((unused)),
 		struct lttng_ust_ring_buffer_channel *chan,
@@ -574,8 +601,83 @@ static int client_instance_id(struct lttng_ust_ring_buffer *buf,
 		struct lttng_ust_ring_buffer_channel *chan __attribute__((unused)),
 		uint64_t *id)
 {
+#ifdef RING_BUFFER_CLIENT_HAS_CPU_ID
 	*id = buf->backend.cpu;
+#else
+	(void) buf;
+	*id = 0;
+#endif
+	return 0;
+}
 
+static int client_packet_initialize(struct lttng_ust_ring_buffer *buf,
+		struct lttng_ust_ring_buffer_channel *chan,
+		void *packet,
+		uint64_t timestamp_begin, uint64_t timestamp_end,
+		uint64_t sequence_number, uint64_t events_discarded,
+		uint64_t *packet_length, uint64_t *packet_length_padded)
+{
+	struct packet_header *packet_header = (struct packet_header *)packet;
+	struct lttng_ust_channel_buffer *lttng_chan_buf;
+	ssize_t page_size = LTTNG_UST_PAGE_SIZE;
+	size_t size;
+
+	assert(packet);
+	assert(packet_length);
+	assert(packet_length_padded);
+	if (!buf || !chan)
+		return -EINVAL;
+	if (page_size < 0)
+		return -EINVAL;
+
+	size = LTTNG_UST_ALIGN(client_packet_header_size(), page_size);
+	memset(packet, 0, size);
+	lttng_chan_buf = channel_get_private(chan);
+	/*
+	 * See client_buffer_begin().
+	 */
+	packet_header->magic = CTF_MAGIC_NUMBER;
+	memcpy(packet_header->uuid, lttng_chan_buf->priv->uuid, sizeof(lttng_chan_buf->priv->uuid));
+	packet_header->stream_id = lttng_chan_buf->priv->id;
+	packet_header->stream_instance_id = buf->backend.cpu;
+	packet_header->ctx.timestamp_begin = timestamp_begin;
+	packet_header->ctx.timestamp_end = timestamp_end;
+	packet_header->ctx.content_size = client_packet_header_size() * CHAR_BIT;
+	packet_header->ctx.packet_size = size * CHAR_BIT;
+	packet_header->ctx.events_discarded = (unsigned long) events_discarded;
+#ifdef RING_BUFFER_CLIENT_HAS_CPU_ID
+	packet_header->ctx.cpu_id = buf->backend.cpu;
+#endif
+	packet_header->ctx.packet_seq_num = sequence_number;
+	*packet_length = client_packet_header_size();
+	*packet_length_padded = size;
+	return 0;
+}
+
+/*
+ * The caller must free the pointer stored in packet when done.
+ */
+static int client_packet_create(void **packet, uint64_t *packet_length)
+{
+	ssize_t page_size = LTTNG_UST_PAGE_SIZE;
+	void *new_packet = NULL;
+	size_t size;
+
+	assert(packet);
+	assert(packet_length);
+
+	if (page_size < 0)
+		return -EINVAL;
+	size = LTTNG_UST_ALIGN(client_packet_header_size(), page_size);
+	new_packet = zmalloc(size);
+	if (!new_packet) {
+		*packet = NULL;
+		*packet_length = 0;
+		return -ENOMEM;
+	}
+
+	*packet = new_packet;
+	*packet_length = size;
 	return 0;
 }
 
@@ -598,9 +700,12 @@ struct lttng_ust_client_lib_ring_buffer_client_cb client_cb = {
 	.content_size = client_content_size,
 	.packet_size = client_packet_size,
 	.stream_id = client_stream_id,
+	.current_events_discarded = client_current_events_discarded,
 	.current_timestamp = client_current_timestamp,
 	.sequence_number = client_sequence_number,
 	.instance_id = client_instance_id,
+	.packet_create = client_packet_create,
+	.packet_initialize = client_packet_initialize,
 };
 
 static const struct lttng_ust_ring_buffer_config client_config = {
@@ -613,10 +718,9 @@ static const struct lttng_ust_ring_buffer_config client_config = {
 	.cb.buffer_finalize = client_buffer_finalize,
 	.cb.content_size_field = client_content_size_field,
 	.cb.packet_size_field = client_packet_size_field,
-
-	.tsc_bits = LTTNG_COMPACT_TSC_BITS,
-	.alloc = RING_BUFFER_ALLOC_PER_CPU,
-	.sync = RING_BUFFER_SYNC_GLOBAL,
+	.timestamp_bits = LTTNG_COMPACT_TIMESTAMP_BITS,
+	.alloc = RING_BUFFER_ALLOC_TEMPLATE,
+	.sync = RING_BUFFER_SYNC_PER_CHANNEL,
 	.mode = RING_BUFFER_MODE_TEMPLATE,
 	.backend = RING_BUFFER_PAGE,
 	.output = RING_BUFFER_MMAP,
@@ -687,12 +791,10 @@ int lttng_event_reserve(struct lttng_ust_ring_buffer_ctx *ctx)
 	struct lttng_ust_ring_buffer_ctx_private *private_ctx;
 	uint32_t event_id;
 
-	event_id = event_recorder->priv->id;
+	event_id = (uint32_t) event_recorder->priv->parent.id;
 	client_ctx.chan_ctx = lttng_ust_rcu_dereference(lttng_chan->priv->ctx);
-	client_ctx.event_ctx = lttng_ust_rcu_dereference(event_recorder->priv->ctx);
 	/* Compute internal size of context structures. */
 	ctx_get_struct_size(ctx, client_ctx.chan_ctx, &client_ctx.packet_context_len);
-	ctx_get_struct_size(ctx, client_ctx.event_ctx, &client_ctx.event_context_len);
 
 	nesting = lib_ring_buffer_nesting_inc(&client_config);
 	if (nesting < 0)
@@ -782,19 +884,30 @@ static
 int lttng_flush_buffer(struct lttng_ust_channel_buffer *chan)
 {
 	struct lttng_ust_ring_buffer_channel *rb_chan = chan->priv->rb_chan;
-	struct lttng_ust_ring_buffer *buf;
-	int cpu;
+	int shm_fd, wait_fd, wakeup_fd;
+	uint64_t memory_map_size;
+	void *memory_map_addr;
 
-	for_each_channel_cpu(cpu, rb_chan) {
-		int shm_fd, wait_fd, wakeup_fd;
-		uint64_t memory_map_size;
-		void *memory_map_addr;
+	if (client_config.alloc == RING_BUFFER_ALLOC_PER_CHANNEL) {
+		struct lttng_ust_ring_buffer *buf;
 
 		buf = channel_get_ring_buffer(&client_config, rb_chan,
-				cpu, rb_chan->handle, &shm_fd, &wait_fd,
+				0, rb_chan->handle, &shm_fd, &wait_fd,
 				&wakeup_fd, &memory_map_size, &memory_map_addr);
 		lib_ring_buffer_switch(&client_config, buf,
 				SWITCH_ACTIVE, rb_chan->handle);
+	} else {
+		int cpu;
+
+		for_each_channel_cpu(cpu, rb_chan) {
+			struct lttng_ust_ring_buffer *buf;
+
+			buf = channel_get_ring_buffer(&client_config, rb_chan,
+					cpu, rb_chan->handle, &shm_fd, &wait_fd,
+					&wakeup_fd, &memory_map_size, &memory_map_addr);
+			lib_ring_buffer_switch(&client_config, buf,
+					SWITCH_ACTIVE, rb_chan->handle);
+		}
 	}
 	return 0;
 }
